#!/usr/local/bin/php -q
<?php

/*
**   Copyright (C) Relog AG, Zrich, Switzerland (http://relog.ch)
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU Library General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
**
**   $Id: phpdoc.php,v 1.18.2.1 2001/04/03 19:15:47 mdaenzer Exp $
**
**   PHP API documentation tool
*/


/* no limits */
set_time_limit(0);



require("lib/phpdoc/layout_box.class");

require("lib/phpdoc/pd_entity.class");
require("lib/phpdoc/pd_class.class");
require("lib/phpdoc/pd_function.class");
require("lib/phpdoc/pd_arg.class");
require("lib/phpdoc/pd_var.class");

/* array holding all classes */
$classes = array();
$current_class = "";

/* array holding all functions */
$functions = array();
$current_function = "";

/* directory where the generated files are written to */
//$directory = $PWD;

/* Regular expression to get the class files */
$suffixes = '\.(class|p?inc|php(3|4)?)$';


/* Debug output */
function debugf($s)
{
  if ($GLOBALS["DEBUG"])
    echo $s;
}


/* Normalize a path, i.e. fold . and .. */
function _normalize_path($path)
{
   $pre = ereg("^/", $path) ? "/" : "";

   $result = array();

   $parts = explode('/', $path);

   for ($i = $j = 0; $i < count($parts); $i++)
   {
      switch ($parts[$i])
      {
      case '': case '.': break;
      case '..': $result[--$j] = ""; if ($j < 0) $j = 0; break;
      default: $result[$j++] = $parts[$i]; break;
      }
   }

   return $pre . ereg_replace('/+$', '', implode('/', $result));
}


/*
 * Parser functions
 */

function startclass($name, $basename, &$info, $dirname)
{
   global $classes, $current_class;

   if ($current_class)
      endclass();

   if ($classes[$name])
      echo "WARNING: class $name already exists - skipping\n";
   else
   {
      $current_class = $name;
      $classes[$name] = new pd_class($name, $basename, $info, $dirname);

      debugf(">>>>class $name\n");
   }

   $info = array();
}

function endclass()
{
   global $current_class;

   if ($current_class)
     debugf("<<<<class $current_class\n");

   $current_class = "";
}

function startfunction($name, &$parameters, &$info)
{
   global $functions, $current_function, $classes, $current_class;

   if ($current_function)
      endfunction();

   $current_function = "$current_class::$name";

   if ($functions[$current_function])
      echo "WARNING: function $current_function already exists - skipping\n";
   else
   {
      if ($classes[$current_class])
         $classes[$current_class]->add_function($name);

      $functions[$current_function] = new pd_function($name, $info, $current_class);

      debugf(">>function $name\n");
   }

   $parameterinfo = $info['@param'];
   reset($parameters);

   while (list(, $parameter) = each($parameters))
   {
      $parameterdescription = $parameterinfo[strip_param($parameter)];

      if (is_array($parameterdescription))
         $parameterdescription = join(" ", $parameterdescription);

      $functions[$current_function]->add_param($parameter, $parameterdescription);

      debugf(">>parameter $parameter:$parameterdescription\n");
   }

   $info = array();
}

function endfunction()
{
   global $current_function;

   if ($current_function)
     debugf("<<function $current_function\n");
}

function storeinfo(&$info, $type, $part, $data)
{
   if (!is_array($info[$type]))
      $info[$type] = array();

   if (!is_array($info[$type][$part]))
      $info[$type][$part] = array();

   $info[$type][$part][] = join(" ", $data);
}

function strip_param($part)
{
   $part = ereg_replace('=.*$', '', $part);
   return ereg_replace('&', '', $part);
}


/**
 * Parse an array of strings representing the source code lines
 * @param $lines Array of strings one per source code line
 * @return Nothing
 * @see parsefiles
 */
function parselines($lines = 3, $dirname)
{
   /* Dynamic array containing collected information from comments */
   $info = array();

   $class = '';
   $baseclass = '';
   $parameters = array();
   $incomment = 0;
   $type = "";
   $part = "";
   $comment = array();

   for ($i = 0; $i < count($lines); $i++)
   {
      $line = $lines[$i];
      $line = trim($line);
      /* Eat one line comments */
      $line = ereg_replace('/\*[^*]*\*/', ' ', $line);
      /* Strip parentheses and commas in function definitions */
      $line = ereg_replace('(function.*)\( *', '\1 ', $line);
      $line = ereg_replace('(\$[^ ,]*) *,* *', '\1 ', $line);
      /* Collapse multiple spaces into one */
      $line = ereg_replace("[ \t]+", " ", $line);
      /* Remove spaces around = in function parameters */
      $line = ereg_replace(' *= *', '=', $line);

      $words = explode(' ', $line);

      switch (strtolower($words[0]))
      {
      case '/**':
         $incomment = 1;
         $comment = array();
         $type = "";
         $part = "";
         /* FALLTHROUGH */

      case '*':
         if (!$incomment)
            break;

         $j = 1;

         if (substr($words[1], 0, 1) == '@')
         {
            storeinfo($info, $type, $part, $comment);
            $type = $words[$j++];

            /* @param have a variable name as first name */
            if ($type == "@param")
            {
               $part = $words[$j++];
               /* XXX */

               debugf("(($part\n");

               $part = strip_param($part);

               debugf("))$part\n");
            }
            else
               $part = "";

            $comment = array();
         }

         for (; $j < count($words); $j++)
         {
            if ($words[$j] == '*/')
            {
               $incomment = 0;
               break;
            }

            $comment[] = $words[$j];
         }

         if ($incomment)
            break;
         /* FALLTHROUGH */

      case '*/':
         $incomment = 0;

         /* Finish comment collection */
         storeinfo($info, $type, $part, $comment);

         $comment = array();
         break;

      case 'class':
         $class = $words[1];
         $baseclass = ($words[2] == 'extends') ? $words[3] : "";
         startclass($class, $baseclass, $info, $dirname);
         break;

      case 'function':
         $function = $words[1];

         $parameters = array();

         for ($p = 2; $p < count($words); $p++)
         {
            if (substr($words[$p], 0, 1) == ')')
               break;

            if (substr($words[$p], -1) == ')')
            {
               $parameters[] = str_replace(')', '', $words[$p]);
               break;
            }

            $parameters[] = $words[$p];
         }

         startfunction($function, $parameters, $info);
         break;
      }

      debugf($baseclass."::".$class."->".$function."[".join("|", $parameters)."] {".$type."=>".join(" ", $comment)."}\n");
   }

   endfunction();
   endclass();
}


/**
 * Parses a file or (recursively) all files if the given file is a directory
 * @param $file File or directory to parse
 */
function parsefiles($file, $dirname="")
{
   global $PWD, $parsed;
   /*
   if (!ereg('^/.*', $file))
     $file = "$PWD/$file";
   */
   /* follow links and normalize path */

   debugf("$file");
   while (is_link($file))
   {
     $target = readlink($file);
     debugf(" => $target");
     if (ereg("^/", $target))
       $file = $target;
     else if (ereg("^\.+/*$", $target))
       /* symlink loop, abort */
       return;
     else
       $file .= "/$target";
   }
   debugf("\n");

   $file = _normalize_path($file);

   /* Recurse into subdirectories */
   if (is_dir($file) && !$parsed[$file])
   {
      $parsed[$file] = true;

      debugf(">>Entering $file...\n");

      ereg("($PWD)?(/)?(.*)(/.*\.html)?", $file, $parts);

      if ($dir = opendir($file))
      {
         while ($subfile = readdir($dir))
         {
            debugf(">>Checking $subfile...\n");

            /* Skip hidden files (especially . and .. ) */
            if (substr($subfile, 0, 1) == '.')
               continue;

            if (is_dir("$file/$subfile") || eregi($GLOBALS["suffixes"], $subfile))
               parsefiles("$file/$subfile", $parts[3]);
         }

         closedir($dir);
      }
   }
   else
   {
      /* Do actual handling of file */
      echo "Processing $file...\n";
      $lines = file($file);
      parselines($lines, $dirname);
   }
}

global $functions, $current_function, $classes, $current_class, $PWD, $parsed;
/*
 * Main
 */

/* prepend current directory's name if given path is relative */
/*
if (!ereg('^/.*', $directory))
   $directory = "$PWD/$directory";
*/
$directory = _normalize_path($directory);

/*
for ($i = 1; $i < $argc; $i++)
{
   parsefiles($argv[$i]);
}
*/
foreach ($files as $val)
{
   parsefiles($val);
}


for ($dirs = array(); list(, $class) = each($classes);)
{
   $dir = $class->dirname;

   if (!is_array($dirs[$dir]))
      $dirs[$dir] = array();

   $dirs[$dir][] = $class->name;
}

$dirnames = array_keys($dirs);
sort($dirnames);


/* Write the HTML files */
$fname = "$directory/index.html";
if ($fd = fopen($fname, "w"))
{
   echo "Writing file $fname...\n";
   fwrite($fd, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\"\"http://www.w3.org/TR/REC-html40/loose.dtd>\n<HTML>\n<HEAD>\n<!-- Generated by phpdoc.php on ".date("D M dS H:i:s Z Y")."-->\n<TITLE>PHP API Documentation</TITLE>\n</HEAD>\n\n<FRAMESET cols=\"30%,70%\">\n<FRAMESET rows=\"40%,60%\">\n<FRAME src=\"index-frame.html\" name=\"indexFrame\">\n<FRAME src=\"start-page.html\" name=\"dirFrame\">\n</FRAMESET>\n<FRAME src=\"start-page.html\" name=\"classFrame\">\n</FRAMESET>\n<NOFRAMES>\n<H2>Frame Alert</H2>\n\n<P>\nThis document is designed to be viewed using the frames feature. If you see this message, you are using a non-frame-capable web client.\n<BR>\nLink to <A HREF=\"index-frame.html\">Non-frame version.</A></NOFRAMES>\n</HTML>");
   fclose($fd);
}
else
   echo "Warning: Couldn't open file $fname for writing!\n";


$fname = "$directory/start-page.html";
if ($fd = fopen($fname, "w"))
{
   echo "Writing file $fname...\n";
   fwrite($fd, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\"\"http://www.w3.org/TR/REC-html40/loose.dtd>\n<HTML>\n\n<BODY BGCOLOR=\"#FFFFFF\">\n<CENTER>Please choose from the left (upper) frame(s).</CENTER>\n</BODY>\n</HTML>");
   fclose($fd);
}
else
   echo "Warning: Couldn't open file $fname for writing!\n";


for (reset($classes); list(, $class) = each($classes);)
{
   echo "Writing file ".($name = "$directory/".$class->get_filename())."...\n";
   $class->write_html_file($name, $classes, $functions);
   $names[] = $class->name;
}


/* class lists for directories */
for (reset($dirnames); list(, $dir) = each($dirnames);)
{
   if ($fd = fopen($fname = "$directory/classes-".strtr($dir,"/","-").".html", "w"))
   {
      echo "Writing file $fname...\n";
      fwrite($fd, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\"\"http://www.w3.org/TR/REC-html40/loose.dtd>\n<HTML>\n<HEAD>\n<!-- Generated by phpdoc.php on ".date("D M dS H:i:s Z Y")."-->\n<TITLE>PHP API Documentation</TITLE>\n</HEAD>\n\n<BODY BGCOLOR=\"#FFFFFF\">\n<BR>\n<H2>Classes in Directory $dir</H2>\n");

      sort($dirs[$dir]);

      while (list(,$name) = each($dirs[$dir]))
         fwrite($fd, "<A HREF=\"".$classes[$name]->get_filename()."\" TARGET=\"classFrame\">$name</A><BR>\n");

      fwrite($fd, "\n</BODY>\n</HTML>");
      fclose($fd);
   }
}


/* helper functions (not class members) */
$fname = "$directory/function-summary.html";
for ($names = array(); list($name,) = each($functions);)
  if (ereg('^::', $name))
    $names[] = $name;

if (count($names))
{
  $helpers = true;

  if ($fd = fopen($fname, "w"))
  {
    echo "Writing file $fname...\n";
    fwrite($fd, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\"\"http://www.w3.org/TR/REC-html40/loose.dtd>\n\n<HTML>\n<HEAD>\n<!-- Generated by phpdoc.php on ".date("D M dS H:i:s Z Y")."-->\n<TITLE>PHP API Documentation - Helper Functions</TITLE>\n</HEAD>\n\n<BODY BGCOLOR=\"#FFFFFF\">\n<BR>\n<H2>Helper Functions</H2>\n<P>\n");

    sort($names);

    /* summary */
    $box = new pd_layout_box($fd, "Summary", 1, 2);
    $box->set_column_width("10\%");
    $box->open();

    for ($first=true; list(, $name) = each($names); $first=false)
    {
      if (!$first) $box->newline();

      $func = $functions[$name];
      $pre = ($func->ret == "") ? "" : "non-";
      fwrite($fd, "<FONT SIZE=\"-2\"><CODE>{$pre}void</CODE></FONT>");
      $box->set_column_width("90\%");
      $box->newfield();
      $func->write_summary($fd, $classes);
      $box->set_column_width("10\%");
    }
    $box->close();

    /* detail */
    $box = new pd_layout_box($fd, "Detail", 0);
    $box->open();
    $box->close();

    for (reset($names); list(, $name) = each($names);)
      $functions[$name]->write_detail($fd, $classes, $functions);

    fwrite($fd, "</BODY>\n</HTML>");
    fclose($fd);
  }
  else
    echo "Warning: Couldn't open file $fname for writing!\n";
}
else
  $helpers = false;


/* list of directories */
$fname = "$directory/index-frame.html";
if ($fd = fopen($fname, "w"))
{
   echo "Writing file $fname...\n";
   fwrite($fd, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\"\"http://www.w3.org/TR/REC-html40/loose.dtd>\n<HTML>\n<HEAD>\n<!-- Generated by phpdoc.php on ".date("D M dS H:i:s Z Y")."-->\n<TITLE>PHP API Documentation</TITLE>\n</HEAD>\n\n<BODY BGCOLOR=\"#FFFFFF\">\n<BR>\n");

   if ($helpers)
     fwrite($fd, "<A HREF=\"function-summary.html\" TARGET=\"classFrame\">Helper functions</A><BR>\n");

   fwrite($fd, "<H2>Directories</H2>\n");

   for (reset($dirnames); list(, $name) = each($dirnames);)
      fwrite($fd, "<A HREF=\"classes-".strtr($name,"/","-").".html\" TARGET=\"dirFrame\">$name</A><BR>\n");

   fwrite($fd, "</BODY>\n</HTML>");
   fclose($fd);
}
else
   echo "Warning: Couldn't open file $fname for writing!\n";


if ($DEBUG)
{
   if (count($classes))
      echo "\nClasses found:\n";

   for (reset($classes); list(, $class) = each($classes);)
   {
      echo $class->name." (parent: ".$class->parentname.")\n";

      if (count($funcs = $class->funcs))
      {
         echo "\tFunctions:\n";

         while (list(,$name) = each($class->funcs))
         {
            $func = $functions["{$class->name}::$name"];
            echo "\t\t".$func->name."(";

            if (count($params = $func->params))
            {
               while (list(,$param) = each($func->params))
               {
                  echo "\n\t\t\t{$param["name"]}: {$param["desc"]}";
               }

               echo "\n\t\t";
            }

            echo ")\n";
         }
      }

      echo "\n";
   }

   echo "\nFunctions found (not belonging to a class):\n";
   for (reset($functions); (list($name, $func) = each($functions)) && ereg("^::.*", $name);)
   {
      echo "$name(";

      if (count($func->params))
      {
         while (list(,$param) = each($func->params))
         {
            echo "\n\t{$param["name"]}: {$param["desc"]}";
         }

         echo "\n\t";
      }

      echo ")\n\n";
   }
}


?>
